#include "TecModule.h"
#include "TecParameters.h"

/**********************************************************************************************************************/

TecModule::TecModule(EntityOwner* owner, const std::string& name, const std::string& description,
    ctk::HierarchyModifier hierarchyModifier, const std::unordered_set<std::string>& tags)
: ctk::ApplicationModule(owner, name, description, hierarchyModifier, tags) {
  assert(checkDescriptorsList());
  createReadWriteMaps();

  // Fill groupCommandMap
  createCommandGroupMaps();

  // TODO: Should this be combined with the implementation of #6567?
  for(auto& desc : descriptors) {
    if(desc.direction != Direction::Read) continue;

    enablePeriodicReadList.emplace_back(ChimeraTK::ScalarPollInput<int>(this, desc.pvName + "_enaPeriodicRead", "",
                                            "Enable periodic read-out of " + desc.pvName),
        desc.value);
  }
}

/**********************************************************************************************************************/
void TecModule::createReadWriteMaps() {

  auto insertIntoMaps = [this](ParameterDescriptor& d) {
    if (d.valueType == ValueType::Int) {

      readIntMap.emplace(d.value, ctk::ScalarOutput<int>{ this, d.pvName + "_rd", d.unit, d.description });

      if (d.direction == Direction::ReadWrite) {
        writeIntMap.emplace( d.value, ctk::ScalarPushInputWB<int>{ this, d.pvName + "_wr", d.unit, d.description });
      }
    } else if (d.valueType == ValueType::Float) {

      readFloatMap.emplace( d.value, ctk::ScalarOutput<float>{ this, d.pvName + "_rd", d.unit, d.description });

      if (d.direction == Direction::ReadWrite) {
        writeFloatMap.emplace( d.value, ctk::ScalarPushInputWB<float>{ this, d.pvName + "_wr", d.unit, d.description });
      }
    }
  };
  std::for_each(descriptors.begin(), descriptors.end(), insertIntoMaps);
}

/**********************************************************************************************************************/

void TecModule::mainLoop() {
  // enableDebug = 1; // -> to temporarily enable debug output by default
  if(enableDebug) std::cout << "Processing TecModule::mainLoop()" << std::endl;

  tecIface = std::make_shared<TecChimeraInterface>(
      triggerCommands, responseStatus, commandMemoryRaw_wr, commandMemoryRaw_rd, true);
  tec = TecController(tecIface);

  std::unordered_set<ChimeraTK::TransferElementID> writeBackAccessors;

  auto inserter = [&writeBackAccessors](auto& elem) { writeBackAccessors.insert(elem.second.getId()); };

  std::for_each(writeIntMap.begin(), writeIntMap.end(), inserter);
  std::for_each(writeFloatMap.begin(), writeFloatMap.end(),inserter);

  // Create TransferElementID to TecParameter maps
  createTransferIdGroupMaps();

  auto group = readAnyGroup();
  while(true) {
    auto id = group.readAny();

    errorStatus = 0;

    if(id == periodicRead.trigger.getId()) {
      readPeriodic();
    }
    else if(readGroupMap.find(id) != readGroupMap.end()) {
      readGroup(readGroupMap.at(id));
    }
    else if(writeGroupMap.find(id) != writeGroupMap.end()) {
      writeGroup(writeGroupMap.at(id));
    }
    else if(writeGroupToFlashMap.find(id) != writeGroupToFlashMap.end()) {
      writeGroupToFlash(writeGroupToFlashMap.at(id));
    }
    else if(setInputFromOutputGroupMap.find(id) != setInputFromOutputGroupMap.end())
      setInputFromOutputGroup(setInputFromOutputGroupMap.at(id));
    else if(writeBackAccessors.count(id) == 0){
      std::cout << "Unexpected input (id = " << id << ") has been written, don't know what to do!" << std::endl;
    }
  }
}

/**********************************************************************************************************************/

void TecModule::readPeriodic() {
  if(enableDebug) std::cout << "readPeroidic()" << std::endl;

  std::list<TecFrame> readRequests;
  for(auto& entry : enablePeriodicReadList) {
    if(entry.first != 0) {
      if(enableDebug) std::cout << "Querying value for " << entry.second << std::endl;
      readRequests.emplace_back(0, entry.second, TecFrame::TecFrameType::TecParameterValueRead, 1);
    }
  }

  auto answers = doTecCommunication(readRequests);

  for(auto frame : answers) {
    if(frame->getState() != TecFrame::TecFrameParsed) {
      std::cout << "Error: Frame not parsed, skipping" << std::endl;
      continue;
    }

    try {
      auto& accessor = readIntMap.at(frame->getParameterOffset());
      accessor = frame->getIntParameterValue();
      accessor.write();

      continue;
    } catch(std::out_of_range&)  {
      // Intentionally left blank, could be float
    }

    try {
      auto& accessor = readFloatMap.at(frame->getParameterOffset());
      accessor = frame->getFloatParameterValue();
      accessor.write();
    } catch(std::out_of_range&)  {
      // FIXME: Once done with mapping, throw exception?
      std::cout << "Unmapped TecParameter " << frame->getParameterOffset() << ", skipping" << std::endl;
    }
  }
}

/**********************************************************************************************************************/

void TecModule::readGroup(std::list<TecFrame::TecParameter>& group) {
  if(enableDebug) std::cout << "readGroup()" << std::endl;

  std::list<TecFrame> readRequests;
  for(auto& param : group) {
    if(enableDebug) std::cout << "Add read frame for param: " << param << std::endl;
    readRequests.emplace_back(0, param, TecFrame::TecFrameType::TecParameterValueRead, 1);
  }

  auto answers = doTecCommunication(readRequests);

  for(auto frame : answers) {
    if(enableDebug) {
      std::cout << "STATE = " << frame->getState() << std::endl;
      frame->print();
    }
    if(frame->getState() == TecFrame::TecFrameState::TecFrameParsed) {
      if(readIntMap.find(frame->getParameterOffset()) != readIntMap.end()) {
        if(enableDebug) {
          std::cout << "Setting integer paramter " << frame->getParameterOffset() << " to "
                    << frame->getParameterValue().integerValue << std::endl;
        }
        auto& accessor = readIntMap.at(frame->getParameterOffset());
        accessor = frame->getParameterValue().integerValue;
        accessor.write();
      }
      else {
        assert(readFloatMap.find(frame->getParameterOffset()) != readFloatMap.end());
        if(enableDebug) {
          std::cout << "Setting float paramter " << frame->getParameterOffset() << " to "
                    << frame->getParameterValue().floatValue << std::endl;
        }
        auto& accessor = readFloatMap.at(frame->getParameterOffset());
        accessor = frame->getParameterValue().floatValue;
        accessor.write();
      }
    }
  }
  if(enableDebug) std::cout << "done." << std::endl;
}

/**********************************************************************************************************************/

void TecModule::writeGroup(std::list<TecFrame::TecParameter>& group) {
  if(enableDebug) std::cout << "writeGroup()" << std::endl;

  // create request to disable write to flash
  std::list<TecFrame> writeRequests;
  writeRequests.emplace_back(0, TecFrame::ParameterSystemFlashSaveOff, TecFrame::TecParameterValueSet, 1, (int)1);

  writeGroupInternal(group, writeRequests);
}

/**********************************************************************************************************************/

void TecModule::writeGroupToFlash(std::list<TecFrame::TecParameter>& group) {
  if(enableDebug) std::cout << "writeGroupToFlash()" << std::endl;

  // create request to enable write to flash
  std::list<TecFrame> writeRequests;
  writeRequests.emplace_back(0, TecFrame::ParameterSystemFlashSaveOff, TecFrame::TecParameterValueSet, 1, (int)0);

  writeGroupInternal(group, writeRequests);
}

/**********************************************************************************************************************/
void TecModule::writeGroupInternal(std::list<TecFrame::TecParameter>& group, std::list<TecFrame>& writeRequests) {
  for(auto& param : group) {
    try {
      auto& accessor = writeIntMap.at(param);
      writeRequests.emplace_back(0, param, TecFrame::TecFrameType::TecParameterValueSet, 1, accessor);
      continue;
    }
    catch(std::out_of_range& ex) {
      // Intentionally left blank
    }

    try {
      auto& accessor = writeFloatMap.at(param);
      writeRequests.emplace_back(0, param, TecFrame::TecFrameType::TecParameterValueSet, 1, accessor);
    }
    catch(std::out_of_range& ex) {
      std::cout << "Unmapped TecParameter " << param << ", skipping" << std::endl;
    }
  }

  auto answers = doTecCommunication(writeRequests);

  for(auto f : answers) {
    if(f->getState() != TecFrame::TecFrameState::TecFrameParsed) continue;

    try {
      auto& accessor = readIntMap.at(f->getParameterOffset());
      accessor = f->getIntParameterValue();
      accessor.write();
      continue;
    } catch(std::out_of_range&)  {
      // Intentionally left blank
    }
    try {
      auto& accessor = readFloatMap.at(f->getParameterOffset());
      accessor = f->getFloatParameterValue();
      accessor.write();
      continue;
    } catch(std::out_of_range&)  {
      std::cout << "Unmapped TecParameter " << f->getParameterOffset() << ", skipping" << std::endl;
    }
  }
}
/**********************************************************************************************************************/

/**
 * Iterates over all ParameterDescriptors and maps the TransferElementID of
 * the related input in groupCommandMap to the TecParameter value
 */
void TecModule::createTransferIdGroupMaps() {
  for(auto descriptor : descriptors){
    try {
      if(!descriptor.readGroup.empty()) {
        readGroupMap[groupCommandMap.at(descriptor.readGroup).getId()].emplace_back(descriptor.value);
      }
      if(!descriptor.writeGroup.empty()) {
        writeGroupMap[groupCommandMap.at(descriptor.writeGroup).getId()].emplace_back(descriptor.value);
      }
      if(!descriptor.writeToFlashGroup.empty()) {
        writeGroupToFlashMap[groupCommandMap.at(descriptor.writeToFlashGroup).getId()].emplace_back(descriptor.value);
      }
      if(!descriptor.setFromReadbackGroup.empty()) {
        setInputFromOutputGroupMap[groupCommandMap.at(descriptor.setFromReadbackGroup).getId()].emplace_back(
            descriptor.value);
      }
    }
    catch(std::out_of_range&) {
      std::cout << "Missing accessor in groupCommandMap! This warning should later become a hard error!" << std::endl;
    }
  }
}

/**********************************************************************************************************************/

std::list<std::shared_ptr<TecFrame>> TecModule::doTecCommunication(std::list<TecFrame>& requests) {
  // No frames passed, nothing to do
  if(requests.size() == 0) return {};

  std::list<std::shared_ptr<TecFrame>> answers;

  // add frames
  tec.clearAllFrames();
  auto requestIt = requests.begin();
  while(true) {
    bool commitTransaction = false;
    if(tec.addFrame(*requestIt)) {
      if(enableDebug) std::cout << "Frame added for parameter " << requestIt->getParameterOffset() << std::endl;
      // frame successfully added
      requestIt = requests.erase(requestIt); // erase from list and increment iterator
      if(requestIt == requests.end()) commitTransaction = true;
    }
    else {
      // frame was not added because maximum number of frames was exceeded
      commitTransaction = true;
      if(enableDebug) std::cout << "Maximum number of frames reached, need to split transaction." << std::endl;
    }
    if(commitTransaction) {
      if(enableDebug) std::cout << "constructAllFrames:" << std::endl;
      tec.constructAllFrames();
      if(enableDebug) {
        for(uint8_t i = 0; i < tec.getFrameCount(); i++) {
          tec.getFrameAt(i)->print();
        }
        std::cout << "copyCommandsToHardware:" << std::endl;
      }
      tec.copyCommandsToHardware();
      if(enableDebug) std::cout << "triggerHardwareCommands:" << std::endl;
      tec.triggerHardwareCommands();

      if(enableDebug) std::cout << "readHardwareResponses:" << std::endl;
      uint32_t msPassed{0};
      bool hasTimeout{false};
      while(not tec.readHardwareResponses()) {

        if(msPassed == fpgaTimeout){
          const std::string timeoutMessage = "Timeout limit for hardware response exceeded!"; 
          if(enableDebug) std::cout << "  " + timeoutMessage << std::endl;

          // TODO: Needs discussion on how to combine with msg implemented
          // in #6571 (serial comm timeout)
          if(errorStatus){
            errorMessage = errorMessage + ", " + timeoutMessage;
          }
          else{
            errorStatus = 1;
            errorMessage = timeoutMessage;
          }

          hasTimeout = true;
          break;  
        }
        usleep(1000);
        ++msPassed;
      }

      // In case of timeout, we don't have a valid response
      if(not hasTimeout){
        if(enableDebug) std::cout << "copyResponsesFromHardware:" << std::endl;
        tec.copyResponsesFromHardware();
        if(enableDebug) std::cout << "parseAllFrames:" << std::endl;
        tec.parseAllFrames();

        for(auto i = 0; i < tec.getFrameCount(); i++) {
          answers.push_back(tec.getFrameAt(i));
        }
      }

      // remove all frames for a fresh start of the next transaction
      tec.clearAllFrames();

      // bail out if at end of request list or timeout
      if(requestIt == requests.end() || hasTimeout) break;
    }
  }
  return answers;
}

/**********************************************************************************************************************/

void TecModule::setInputFromOutputGroup(std::list<TecFrame::TecParameter>& group) {
  for (auto& param : group) {
    if (readIntMap.find(param) != readIntMap.end()) {
      auto& readAccessor = readIntMap.at(param);
      assert(writeIntMap.find(param) != writeIntMap.end());
      auto& writeAccessor = writeIntMap.at(param);

      writeAccessor = static_cast<int>(readAccessor);
      writeAccessor.write();

      if (enableDebug) {
        std::cout << "Copying int value from " << readAccessor.getName() << " to " << writeAccessor.getName()
                  << std::endl;
      }
    } else {
      assert(readFloatMap.find(param) != readFloatMap.end());
      assert(writeFloatMap.find(param) != writeFloatMap.end());

      auto& readAccessor = readFloatMap.at(param);
      auto& writeAccessor = writeFloatMap.at(param);

      writeAccessor = static_cast<float>(readAccessor);
      writeAccessor.write();

      if (enableDebug) {
        std::cout << "Copying float value from " << readAccessor.getName() << " to " << writeAccessor.getName()
                  << std::endl;
      }
    }
  }
}

/**********************************************************************************************************************/

/**
 * Iterate over descriptions and maps group names to their respective trigger command accessors
 */
void TecModule::createCommandGroupMaps()
{
  for(auto& d : descriptors) {
    for(auto& group : {d.readGroup, d.writeGroup, d.writeToFlashGroup, d.setFromReadbackGroup}) {
      if(group.empty()) continue;
      groupCommandMap.emplace(group, ChimeraTK::ScalarPushInput<int>{this, group + "_wr", "", "Trigger command"});
    }
  }
}
