#include "TecDummy.h"
#include <arpa/inet.h>

#include <tec/TecFrame.h>

#include <boost/algorithm/string.hpp>
#include <sstream>

TecDummy::BackendRegisterer TecDummy::backendRegisterer;

/*********************************************************************************************************************/

TecDummy::BackendRegisterer::BackendRegisterer() {
  std::cout << "TecDummy::BackendRegisterer: registering backend type TecDummy" << std::endl;
  ChimeraTK::BackendFactory::getInstance().registerBackendType("TecDummy", &TecDummy::createInstance);
}

/*********************************************************************************************************************/

void TecDummy::read(uint8_t bar, uint32_t address, int32_t* data, size_t sizeInBytes) {
  DummyBackend::read(bar, address, data, sizeInBytes);

  if(bar != TecDummy_bar || address != TecDummy_address_loopDone) return;
  if(reg_loopStart != 1) return;

  // convert command into string
  // FIXME: this can be done better if the DummyRegisterRawAccessor would allow us to use the raw pointer to this...
  std::string commandString(reg_command.getNumberOfElements() * 4, 0);
  for(size_t i = 0; i < reg_command.getNumberOfElements(); ++i) {
    uint32_t word = ntohl(reg_command[i]);
    std::memcpy(&(commandString[4 * i]), &word, 4);
  }

  // process frame by frame
  constexpr size_t nFrames = 16;
  constexpr size_t frameLength = 16 * 4;
  std::string reply;
  for(size_t frame = 0; frame < nFrames; ++frame) {
    size_t start = frame * frameLength + 8;
    auto frameData = commandString.substr(start, frameLength);

    // remove everything from the 0x0D character, because only 0x00 following
    auto endOfFrame = frameData.find_first_of("\r");
    if(endOfFrame == std::string::npos) {
      reply += std::string(frameLength, '\0');
      continue;
    }
    frameData.erase(endOfFrame);

    // perform data processing
    auto frameReply = processFrame(frameData);
    boost::to_upper(frameReply);

    // append to answer
    frameReply += "\r";
    frameReply += std::string(frameLength - frameReply.size(), '\0');
    reply += frameReply;
  }
  assert(reply.size() == nFrames * frameLength);

  // convert reply back into byte stream
  for(size_t i = 0; i < reply.length() / 4; ++i) {
    uint32_t word;
    std::memcpy(&word, &(reply[4 * i]), 4);
    reg_command[i] = htonl(word);
  }

  *data = 1;
}

/*********************************************************************************************************************/

std::map<TecDummy::TecFramePart, std::string> TecDummy::splitFrame(const std::string& command) {
  std::map<TecDummy::TecFramePart, std::string> result{};
  int offset = 0;
  // Reqeust or response marker
  result[TecFramePart::MARKER] = command.substr(offset, 1);
  offset++;

  // TEC address
  result[TecFramePart::ADDRESS] = command.substr(offset, 2);
  offset += 2;

  // Sequence ID
  result[TecFramePart::SEQUENCE] = command.substr(offset, 4);
  offset += 4;

  // Command
  if(command[offset] == '?') {
    result[TecFramePart::COMMAND] = command.substr(offset, 3);
    offset += 3;
  }
  else {
    result[TecFramePart::COMMAND] = command.substr(offset, 2);
    offset += 2;
  }

  // Parameter ID
  result[TecFramePart::PARAMETER_ID] = command.substr(offset, 4);
  offset += 4;

  // Channel ID
  result[TecFramePart::CHANNEL] = command.substr(offset, 2);
  offset += 2;

  // payload
  auto residual_length = command.length() - offset - 4;

  if(residual_length > 0) {
    result[TecFramePart::PAYLOAD] = command.substr(offset, residual_length);
    offset += residual_length;
  }

  // CRC
  result[TecFramePart::CRC] = command.substr(offset, 4);

  return result;
}

std::string TecDummy::getCrc(std::string&) {
  return std::string{"DEAD"};
}

std::string TecDummy::processFrame(const std::string& command) {
  std::cout << "TecDummy::processFrame() command = " << command << std::endl;
  auto parsedFrame = splitFrame(command);
  std::string result = "";

  // extract parameter ID
  auto parameterId =
      static_cast<TecFrame::TecParameter>(std::stoul("0x" + parsedFrame[TecFramePart::PARAMETER_ID], nullptr, 16));

  // Handle store command
  if(parsedFrame[TecFramePart::COMMAND] == "VS") {
    int32_t value = std::stol("0x" + parsedFrame[TecFramePart::PAYLOAD], nullptr, 16);

    std::cout << "Storing raw value " << value << "(" << parsedFrame[TecFramePart::PAYLOAD] << ") for parameter " << parameterId << std::endl;

    rawValues[parameterId] = value;
    result = "!" + parsedFrame[TecFramePart::ADDRESS] + parsedFrame[TecFramePart::SEQUENCE];
    result += getCrc(result);
  }
  else if(parsedFrame[TecFramePart::COMMAND] == "?VR") {
    result = "!" + parsedFrame[TecFramePart::ADDRESS] + parsedFrame[TecFramePart::SEQUENCE];
    std::stringstream ss;
    ss << std::hex << std::setw(8) << std::setfill('0') << rawValues[parameterId];
    result += ss.str();
    result += getCrc(result);
  }

  //
  // mark received parameter as written to flash or not
  auto isLastWriteToFlash = (not rawValues.at(TecFrame::ParameterSystemFlashSaveOff)) ? true : false;
  lastWriteToFlash[parameterId] = isLastWriteToFlash;

  std::cout << "Returning result " << result << std::endl;

  return result;
}

/*********************************************************************************************************************/
